/* Copyright (C) Richard Neill 2006-2011, <ftdi245io at REMOVE.ME.richardneill.org>. This program is Free Software. You can
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
 * For the details, please see: http://www.gnu.org/licenses/gpl.html */

/* Note: it was originally written for > 1 port, and has been simplified. See version 1.0 for the original version.) */
/* NOTE: my usage aims to be as consistent as possible with the API docs here: http://www.intra2net.com/de/produkte/opensource/ftdi/documentation/group__libftdi.html */
/* NOTE: this does not cleanly cope when it locks or races against itself. Concurrent invocations do very bad things, even when they SHOULD be OK (eg one read, and another write). Fixed with flock(). */

#include "ftdi245io.h"		/* My configuration header */
#include <ftdi.h>		/* Libftdi, from http://www.intra2net.com/de/produkte/opensource/ftdi/ */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>

void printhelp(){
	fprintf(stderr,"\n"
			"DESCRIPTION:\n"
			"	ftdi245io reads and writes to the FTDI245 Module, named Port A.\n"
			"\n"
			"	  * PortA is an 8-bit input/output port.  (Bitbang mode, output bitmask is 0x%02X)\n"
			"\n"
			"USAGE:  Any ONE of these options may be used:\n"
			"\n"
			"       -h  --help              Print this usage.\n"
			"       --info                  Print info about the version and configuration settings used.\n"
			"       --check                 Check for FTDI245s, and check that the expected module is connected.\n"
			"\n"
			"       --check_reset           Do --check; also reset the FTDI device and purge the internal buffers.\n"
			"                               (This exits as soon as it encounters any device which it can't reset/purge.)\n"
			"\n"
			"       --porta_setpins 0xXX    Write (masked) byte 0xXX to port A. Returns immediately, then exits.\n"
			"                               Specify the byte as Hex (e.g. 0xFF), Decimal (e.g. 255), or Octal (e.g. 0377)\n"
			"                               The Output Mask is 0x%02X; input-pins will not be changed.\n"
			"\n"
			"       --porta_readpins        Read (masked) byte from the input pins of port A. This is written to stdout as\n"
			"                               a single byte in the form 0xXX\\n. The program then exits. The Input Mask is 0x%02X.\n"
			"\n"
			"       --porta_readback        Read back the currently buffered byte from port A. We can read back the byte to \n"
			"                               get the current state of the outputs. This is written to stdout as a single byte\n"
			"                               in the form 0xXX\\n. The program then exits. \n"
			"\n"
			"       --porta_test            Write a test pattern to port A. This pattern consists of 12 seconds\n"
			"                               (128 cycles) of 10Hz \"Running LEDs\". The LEDs move from LSB -> MSB. \n"
			"                               Port A's \"carriers\" are \"electrons\", rather than holes.\n"
			"                               At the end of the test, 0x0A remains on (to identify it).\n"
			"                               NOTE: the output mask for PortA is *ignored* here.\n"
			"\n"
			"       --porta_identify_bit N  Flash bit N, (where N is 0..7) at 2Hz, for identification purposes\n"
			"                               To check phase, it spends 0.125s high, 0.375s low\n"
			"                               This never exits: use Ctrl-C to stop it.\n"
			"                               NOTE: the output mask for PortA is *ignored* here.\n"
			"\n"
			"       --porta_zero            Set outputs on port A to zero. Useful to prevent parasitically powering\n"
			"                               the rest of the circuit during a power-on reset. Equivalent to --porta_setpins 0x00.\n"
			"\n"
			"REQUIREMENTS:\n"
			"       You may need to be root, to access the FTDI devices. (Sometimes 'pam_console_apply' helps) \n"
			"       This uses (userspace) libusb, which can automatically dissociate the device from the module without blacklisting.\n"
			"       Allow the bus/kernel up to 5 seconds to settle after plugging in a new device.\n"
			"       If this is slow, and using a USB2 Hub, ensure it has one transaction-translator for each ftdi device, and not just one.\n"
			"       [As of 2007, most generic USB2 hubs do not meet this requirement, but share a single T.T. between all ports.]\n"
			"       If weird things happen, it is often a symptom of insufficient power/excess current-draw.\n"
			"\n"
			"OUTPUTS:\n"
			"       This program always separates stdout from stderr.\n"
			"       The only thing that is written to stdout is the data from --porta_readpins.\n"
			"       stderr receives a confirmation message of what has been done; error messages if failure; and\n"
			"       lots of debugging info if compiled with #define DEBUG 1.\n"
			"\n"
			"EXIT STATUS:\n"
			"       Exit values are defined in ftdi245io.h. %d always means success; nonzero means failure.\n"
			"\n"
			"MORE INFO:\n"
			"       Invoke with --info, or read the source (ftdi245io.c) and header file (ftdi245io.h). Or see man ftdi245io.\n"
			"       [Note that the serial number of the DLP-module is hardcoded in ftdi245io.h; change it there if needed.]\n"
			"\n",
			PORTA_OUTPUT_MASK, PORTA_OUTPUT_MASK, (0xFF & ~PORTA_OUTPUT_MASK), FTDI_EXIT_OK);
}


int main(int argc, char *argv[]){

	int doinfo=0, docheck=0, docheck_reset=0, doporta_setpins=0, doporta_zero=0, doporta_readpins=0, doporta_readback=0, doporta_test=0, doporta_identify=0;	/* Flags: what to do */
	unsigned long input_tmp;
	unsigned char outbyte=0;		/* Output data byte to write */
	unsigned char inbyte=0;			/* Input byte to read */
	unsigned char rawbyte=0;		/* Raw copy (without mask) */
	unsigned char iomask;			/* Output bit mask. 1 = output; 0 = input */
	int ret;				/* Function retval */
	int numdevices;				/* Number of devices found */
	char *porta_found="Missing!";
	int isok=0;				/* All ok? */
	char manufacturer[INFOSIZE], description[INFOSIZE], serial[INFOSIZE];  /* Info about the device */
	char *portname, *portserial;		/* Port name and serial */
	int i;
	char systemstring[512];			/*String for system call command*/
	int lockfd;				/* File descriptor, for locking */

	struct ftdi_context ftdic;		/* FTDI context struct. Usually, accessed by address, i.e. &ftdic */
	struct ftdi_context ftdic_a;		/* Dedicated contexts for Port A */

	struct ftdi_device_list* devlist;	/* List of all ftdi devices */


	/* Get and check the args. */
	if (argc == 2){   /* 1 argument */
		/* Print help, if invoked with -h, or --help */
		if ((!strcmp (argv[1], "-h")) || (!strcmp (argv[1], "--help"))){
			printhelp();
			exit (FTDI_EXIT_OK);

		/* Prepare to print settings info, if invoked with "--info" */
		}else if (!strcmp (argv[1], "--info")){
			doinfo=1;

		/* Prepare to do check, if invoked with "--check" */
		}else if (!strcmp (argv[1], "--check")){
			docheck=1;

		/* Prepare to do check_reset, if invoked with "--check_reset" */
		}else if (!strcmp (argv[1], "--check_reset")){
			docheck_reset=1;

		/* Prepare to do porta_readpins, if invoked with "--porta_readpins" */
		}else if (!strcmp (argv[1], "--porta_readpins")){
			doporta_readpins=1;

		/* Prepare to do porta_readback, if invoked with "--porta_readback" */
		}else if (!strcmp (argv[1], "--porta_readback")){
			doporta_readback=1;

		/* Prepare to do porta_zero, if invoked with "--porta_zero" */
		}else if (!strcmp (argv[1], "--porta_zero")){
			doporta_zero=1;

		/* Prepare to do porta_test, if invoked with "--porta_test" */
 		}else if (!strcmp (argv[1], "--porta_test")){
 			doporta_test=1;

 		}else{
 			fprintf (stderr,"Error: wrong arguments! Use -h for help\n");
 			exit (FTDI_ERROR_WRONGARGS);
 		}

	}else if (argc == 3){	/* 2 arguments were given */

		/* Prepare to do porta_setpins, if invoked with "--porta_setpins" */
		if (!strcmp (argv[1], "--porta_setpins")){
			doporta_setpins=1;

		/* Prepare to do porta_identify_bit, if invoked with "--porta_identify_bit" */
 		}else if (!strcmp (argv[1], "--porta_identify_bit")){
 			doporta_identify=1;

		}else{
			fprintf (stderr,"Error: wrong arguments! Use -h for help\n");
			exit (FTDI_ERROR_WRONGARGS);
		}

		/* Get the output byte to write */
		input_tmp = strtoul(argv[2],NULL,0);
		if (input_tmp > 255){
			fprintf (stderr,"Error: output byte (%s) is out of range.\n",argv[2]);
			exit (FTDI_ERROR_WRONGARGS);
		}else{
			outbyte=input_tmp;
		}

		/* If identifying a bit, get the  relevant byte for this bit */
		if (doporta_identify){
			if (outbyte > 7){
				fprintf (stderr,"Error: output bit to identify (%s) is out of range. Must be 0..7 .\n",argv[2]);
				exit (FTDI_ERROR_WRONGARGS);
			}
			rawbyte = 0x1 << outbyte;
			eprintf ("Flashing bit %s, i.e. byte 0x%x.\n", argv[2], rawbyte);
		}

	}else{	/* Wrong number of args */
		fprintf (stderr,"Error: wrong number of arguments! Need either 1 or 2.\nUse -h for help\n");
		exit (FTDI_ERROR_WRONGARGS);
	}

	/* By now, exactly ONE of these variables is non-zero: doinfo, docheck, doporta_setpins, doporta_readpins, doporta_readback, doporta_test. */


/* Do Info. Print the Module/Port <-> Serial mappings, and the Vendor/Product info. Explain where these come from, and how to change them */
	if (doinfo){
		fprintf(stderr,"\n"
				"The current port mappings for the FTDI245 Modules are as follows:\n"
				"     MODULE   |   SERIAL\n"
				"     ---------------------\n"
				"     PortA    |   %s\n"
				"\n"
				"PortA has IOMask = 0x%02X (1 makes the bit direction an output, 0 is an input).\n"
				"\n"
				"All the modules have VENDOR=0x%X and PRODUCT=0x%X\n"
				"\n"
				"The serial numbers can be read via `lsusb -vvv`, or `ftdi_eeprom`.\n"
				"  e.g. `sudo lsusb -vv -d 0403:6001 | grep iSerial`.\n"
				"The serial numbers may be changed in hardware with `ftdi_eeprom`.\n"
				"The serial numbers we expect are defined in ftdi254io.h\n"
				"\n"
			#if DEBUG == 1
				"* Debugging is enabled; stderr will be verbose.\n"
			#else
				"* Debugging is disabled in ftdi245io.h; stderr will be fairly terse.\n"
			#endif

				"* FTDI USB Latency timer is %d ms. Timeout timer is %d ms.\n"
				"* Buffer size for hardware I/O is %d Bytes. Pipe-capacity is \n"
				"  4096 or 65536 bytes depending on kernel; see man(7) pipe.\n"
				"\n"
				"Written by %s <%s>.\n"
				"This is Free Software; the license is %s.\n"
				"Version %s, Compiled on %s. \n"
				"\n",
				PORTA_SERIAL, PORTA_OUTPUT_MASK, VENDOR, PRODUCT, LATENCY_TIMER, TIMEOUT_TIMER, BUFSIZE,
				AUTHOR, AUTHOR_EMAIL, LICENSE, VERSION, __DATE__);
		exit (FTDI_EXIT_OK);
	}




/* Do Check. Test that all 4 of the devices are present, and can be opened. Note: don't necessarily exit on first failure.*/
	if ((docheck) || (docheck_reset)){

		/* NOTE: flock() omitted here for brevity; see below if needed. */

		/* Initialise ftdi context */
		eprintf("Initialising ftdi...\n");
		ret=ftdi_init(&ftdic);
		if (ret!=0){
			fprintf(stderr, "Error %d: ftdi_init failed (%s)\n", ret, ftdi_get_error_string(&ftdic));
			exit (FTDI_ERROR_INITFAIL);
		}

		/* Get a list of the FTDI devices present. */
		eprintf("Counting FTDI devices...\n");
		ret=ftdi_usb_find_all (&ftdic, &devlist, VENDOR, PRODUCT);
		if (ret<0){
			fprintf(stderr, "Error %d: ftdi_usb_find_all failed (%s)\n", ret, ftdi_get_error_string(&ftdic));
			exit (FTDI_ERROR_OPENFAIL);
		}else{
			numdevices=ret;
			fprintf(stderr, "Number of FTDI devices found: %d\n",ret);
			eprintf("Getting details...\n");

			/* Now, iterate over all the ones actually present, getting their info */
			for (i=0;i<numdevices;i++){

				fprintf(stderr, "Checking device %d of %d...\n",i+1,numdevices);  /* count 1,2,3,4 NOT 0,1,2,3 */
				/* Do NOT call ftdi_usb_open_dev here. If we do, ftdi_usb_get_strings will crash at runtime with a double-free error */

				ret=ftdi_usb_get_strings (&ftdic, devlist->dev, manufacturer, sizeof(manufacturer), description, sizeof(description), serial, sizeof(serial));
				if (ret<0){
					fprintf(stderr, "Error %d: ftdi_usb_get_strings failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
					/* Don't exit here; continue */
				}else{
					/* Match a port if we can */
					if (!strcmp (serial, PORTA_SERIAL)){
						portname="PortA";
						porta_found="Found";
					}else{
						portname="No port defined for this module";
					}
					fprintf(stderr, "Manufacturer='%s', Description='%s', Serial='%s', Port='%s'.\n",manufacturer, description, serial, portname);
				}

				/* Open the device, reset it, and purge its buffers */
				if (docheck_reset){

					eprintf("Opening...\n");
					ret=ftdi_usb_open_dev (&ftdic, devlist->dev);
					if (ret<0){
						fprintf(stderr, "Error %d: ftdi_usb_open_dev failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
						exit(FTDI_ERROR_OPENFAIL);
					}else{
						eprintf("Doing USB Reset...\n");
						ret=ftdi_usb_reset(&ftdic);
						if (ret<0){
							fprintf(stderr, "Error %d: ftdi_usb_reset failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
							exit(FTDI_ERROR_RESETFAIL);
						}

						eprintf("Purge buffers...\n");
						ret=ftdi_usb_purge_buffers(&ftdic);
						if (ret<0){
							fprintf(stderr, "Error %d: ftdi_usb_purge_buffers failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
							exit(FTDI_ERROR_PURGEFAIL);
						}

						eprintf("Close...\n");
						ret=ftdi_usb_close(&ftdic);
						if (ret<0){
							eprintf("Error %d: ftdi_usb_close failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
							exit(FTDI_ERROR_CLOSEFAIL);
						}
						fprintf(stderr, "Device %d has been successfully opened, reset, purged and closed.\n",i+1);
					}
				}

				fprintf(stderr,"\n");

				if (i!=numdevices-1){		/* Only get next device pointer if it exists in the list */
					devlist=devlist->next;
				}
			}
			if (numdevices==0){
				eprintf("[No devices detected]\n\n");
			}
		}

		/* Checking for named ports/modules */
		/* We could do it the hard way, i.e. repeatedly do the following:
		 *	fprintf(stderr, "   * Checking for PortA...");
		 *	ret=ftdi_usb_open_desc(&ftdic, VENDOR, PRODUCT, NULL, PORTA_SERIAL);
		 *	if (ret<0){
	         *		fprintf(stderr, "\n  Error %d: ftdi_usb_open_desc failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
		 *	}else{
		 *	        fprintf(stderr, "(OK). (serial: %s)\n",PORTA_SERIAL);
		 *		ret=ftdi_usb_close(&ftdic);
		 *	if (ret<0){
	         *		fprintf(stderr, "   Error %d: ftdi_usb_close failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
		 *	}
		 * But it is unnecessary, since we have already iterated above over all the devices that do exist.
		 */

		fprintf(stderr,"Checking for named ports...\n");
		fprintf(stderr,"  * PortA (serial: %s)   %s\n",PORTA_SERIAL,porta_found);
		fprintf(stderr,"\n");

		/* Do we have all 4 that we expect?  Don't use numdevices; we want to make sure we have the specific 4 ports, not just any 4 */
		if ( !strcmp(porta_found,"Found") ){
			isok=1;
		}

		/* Free the ftdi device list. */
		/* Note: if there are no devices, this segfaults in libftdi-0.9. The bug has now been fixed and a patch is available. Workaround: only call this if numdevices >0.  */
		ftdi_list_free (&devlist);

		/* De-initialise ftdi context. (void) */
		ftdi_deinit(&ftdic);

		if (isok){
      			/* Exit: we're happy */
			fprintf(stderr, "Found all expected ports.\n");
			exit (FTDI_EXIT_OK);
		}else{
			/* Exit: we're unhappy */
			fprintf(stderr, "Error: one or more of the devices/ports expected cannot be found, or cannot be opened.\n"
					"Check: permissions, devices plugged in, USB has sufficient power, correct serial_numbers in ftdi245io.h ?\n");
			exit (FTDI_ERROR_OPENFAIL);
		}
	}



/* Do output of a single byte, outbyte, to Port A. OR do input of a single byte inbyte, from PortA. */
	if ( (doporta_setpins) || (doporta_readpins) || (doporta_readback) ){

		portname="PortA";
		portserial=PORTA_SERIAL;
		iomask=PORTA_OUTPUT_MASK;

		/* Lock the port. Otherwise, if we race against ourself, horrid things can happen. */
		lockfd = open (PORTA_LOCKFILE, O_RDWR|O_CREAT, S_IRWXU);  /* this mode seems necessary */
		if (lockfd == -1){
			fprintf (stderr, "Error: could not open lockfile %s: %s\n", PORTA_LOCKFILE, strerror(errno));
		}
		ret = flock (lockfd, LOCK_EX | LOCK_NB);  /* Wait for an exclusive lock. (If we have to wait, warn first). */
		if (ret == -1 && errno == EWOULDBLOCK){
			fprintf (stderr, "Note: lockfile %s is currently in use. Waiting till device %s is available.\n", PORTA_LOCKFILE, portname);
			ret = flock (lockfd, LOCK_EX);
		}
		if (ret != 0){
			fprintf (stderr, "Error: failed to perform locking on lockfile %s: %s\n", PORTA_LOCKFILE, strerror(errno));
			exit (FTDI_ERROR_GENERIC);
		}

		/* Initialise ftdi context */
		eprintf("Initialising ftdi context...\n");
		ret=ftdi_init(&ftdic);
		if (ret!=0){
			fprintf(stderr, "Error %d: ftdi_init failed for %s (%s)\n", ret, portname, ftdi_get_error_string(&ftdic));
			exit (FTDI_ERROR_INITFAIL);
		}

		/* Open Port A */
		eprintf("Opening %s (serial: %s)...\n",portname,portserial);
		ret=ftdi_usb_open_desc(&ftdic, VENDOR, PRODUCT, NULL, portserial);
		if (ret<0){
	        	fprintf(stderr, "Error %d: ftdi_usb_open_desc failed for port %s: (%s)\n", ret, portname, ftdi_get_error_string(&ftdic));
			exit (FTDI_ERROR_OPENFAIL);
		}else{

			/* Enable bitbang mode; set output-mask to 0xFF (all outputs), or PORTA_OUTPUT_MASK as appropriate */
			eprintf("Setting %s to bitbang mode, with Output bitmask 0x%02X...\n",portname,iomask);
			ret = ftdi_set_bitmode (&ftdic, iomask, BITMODE_BITBANG);

			if (ret<0){
	        		fprintf(stderr, "Error %d: ftdi_set_bitmode (BITMODE_BITBANG) failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
				exit (FTDI_ERROR_CONFIGFAIL);
			}

			/* Are we writing or reading? */
			if ( (doporta_setpins) ){  /* Writing */

				rawbyte = outbyte; /* save unmasked copy */

				/* Check that outbyte only tries to set the allowed bits designated as outputs in iomask. If not, warn, AND apply the mask */
				if ( (outbyte & iomask) != outbyte ){
					fprintf(stderr, "Warning: output mask for %s is 0x%02X, but you are trying to write the byte 0x%02X. Pins which are designated as inputs will not be changed.\n", portname, iomask, outbyte);
					outbyte = outbyte & iomask;  /* Apply the mask */
				}

				/* Write the byte */
				eprintf("Writing data byte 0x%02X to %s...\n", outbyte,portname);
				ret=ftdi_write_data (&ftdic, &outbyte, 1);	/* Write length is 1 byte */

				if (ret<0){
					fprintf(stderr, "Error %d: ftdi_write_data failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
					exit (FTDI_ERROR_WRITEFAIL);
				}

			}else{		/* Reading. */

				eprintf ("Reading input pins from %s...\n",portname);
				ret=ftdi_read_pins(&ftdic, &inbyte);

				if (ret<0){
					fprintf(stderr, "Error %d: ftdi_read_pins failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
					exit (FTDI_ERROR_READFAIL);
				}

				rawbyte = inbyte; /* save unmasked copy */

				if (doporta_readpins) {		/* Reading real data from PortA */
					/* Mask off the output bits, leaving only the input bits */
					eprintf ("The pin state of %s is 0x%02X. After masking by input mask 0x%02X, it is 0x%02X.\n", portname, inbyte, ~iomask, inbyte & ~iomask);
					inbyte = inbyte & ~iomask;
				}else{				/* Reading back data from PortA's output buffer. */
					/* Don't mask anything; we want the *previous* output bits */
					eprintf ("The pin state of %s is 0x%02X. We are reading back the output buffer, so no input mask is applied.\n", portname, inbyte);
					inbyte = inbyte;    	/* Don't apply the iomask with  "inbyte = inbyte & ~iomask;"  */
				}

				/* Write the data to stdout, followed by a \n */
				printf("0x%02X\n",inbyte);
			}


			/* Do NOT disable bitbang mode, or the written byte won't persist at the outputs. */
			eprintf("NOT disabling bitbang mode for %s, so that the last byte written persists at the outputs.\n",portname);
			/*
			 * eprintf("Disabling bitbang mode for %s...",portname);
			 * ret=ftdi_disable_bitbang (&ftdic);	NOTE: now deprecated anyway, use  ftdi_set_bitmode (&ftdic_a, 0xFF, BITMODE_RESET)  instead.
			 * if (ret<0){
			 *	fprintf(stderr, "\nError %d: ftdi_disable_bitbang failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
			 *	exit (FTDI_ERROR_CONFIGFAIL);
			 * }else{
			 *	eprintf("(OK).\n");
			 * }
			*/

			eprintf("Closing %s...\n",portname);
			ret=ftdi_usb_close(&ftdic);	/* Only try to close if was open; else would segfault*/
			if (ret<0){
	        		fprintf(stderr, "Error %d: ftdi_usb_close failed: (%s)\n", ret, ftdi_get_error_string(&ftdic));
				/* Don't make this a fatal exit, it can (seriously) emit this error!:  "Error -1: ftdi_usb_close failed: (all fine)"  */
				/* exit (FTDI_ERROR_CLOSEFAIL); */
			}
		}

		/* De-initialise ftdi context. (void) */
		ftdi_deinit(&ftdic);

		/* Clean up, unlock. */
		flock (lockfd, LOCK_UN);
		close (lockfd);
		unlink (PORTA_LOCKFILE);

		/* If we got here, it was successful */
		if ( doporta_setpins){
			fprintf(stderr,"Written byte 0x%02X to %s (raw: 0x%02X, iomask: 0x%02X).\n",outbyte,portname, rawbyte, iomask);   /* Say one thing even in non-debug mode */
		}else{
			fprintf(stderr,"Read byte 0x%02X from %s (raw: 0x%02X, iomask: 0x%02X).\n",inbyte,portname, rawbyte, iomask);
		}
		exit (FTDI_EXIT_OK);
	}



/* Zero PortA */
	if (doporta_zero){
		//Rather ugly quick hack: just call self. Otherwise, we'd have to re-factor much of the above.
		eprintf ("Executing %s --porta_setpins 0x%02X...\n",argv[0],0x00);
		sprintf (systemstring, "%s --porta_setpins 0x%02X",argv[0],0x00);
		ret = system (systemstring);
		if (ret != FTDI_EXIT_OK){	//Did it work? If not, exit with the return code from the last instance. (-1 could also mean the system() call failed.
			exit (ret);
		}
		fprintf(stderr,"All the output pins of port A are now low.\n");   /* Say one thing even in non-debug mode */
		exit (FTDI_EXIT_OK);
	}



/* Output a test pattern to Port A.  NOTE: we ignore PORTA_OUTPUT_MASK here. */
	if ((doporta_test) || (doporta_identify)){

		/* NOTE: flock() omitted here for brevity; see above if needed. */

		/* Initialise ftdi contexts */
		eprintf("Initialising ftdi contexts...\n");
		ret=ftdi_init(&ftdic_a);
		if (ret!=0){
			fprintf(stderr, "Error %d: ftdi_init failed for PortA (%s)\n", ret, ftdi_get_error_string(&ftdic_a));
			exit (FTDI_ERROR_INITFAIL);
		}

		/* Open Port A */
		eprintf("Opening PortA (serial: %s)...\n",PORTA_SERIAL);
		ret=ftdi_usb_open_desc(&ftdic_a, VENDOR, PRODUCT, NULL, PORTA_SERIAL);
		if (ret<0){
	        	fprintf(stderr, "Error %d: ftdi_usb_open_desc failed for PortA: (%s)\n", ret, ftdi_get_error_string(&ftdic_a));
			exit (FTDI_ERROR_OPENFAIL);

		}else{

			/* Enable bitbang mode; set output-mask to 0xFF (all outputs) */
			eprintf("Setting PortA to bitbang mode, with Output mask 0xFF (all outputs)...\n");
			if ( PORTA_OUTPUT_MASK != 0xFF ){
				eprintf("NOTE: PortA normally has output mask = 0x%02x. This setting will be *ignored* here.\n",PORTA_OUTPUT_MASK);
			}
			ret = ftdi_set_bitmode (&ftdic_a, 0xFF, BITMODE_BITBANG);

			if (ret<0){
				fprintf(stderr, "Error %d: ftdi_set_bitmode (BITMODE_BITBANG) failed for PortA: (%s)\n", ret, ftdi_get_error_string(&ftdic_a));
				exit (FTDI_ERROR_CONFIGFAIL);
			}

			if (doporta_identify){
				/* Infinite loop, flashing this bit */
				fprintf (stderr, "Flashing bit %s, i.e. byte 0x%x. (0.125s high, 0.375s low).\nCtrl-C to stop.\n", argv[2], rawbyte);
				eprintf("    PortA\n    -----\n");

				while (1){

					/* The relevant bit */
					outbyte = rawbyte;
					eprintf("    %02hhx\n", outbyte);

					/* Write the byte to PortA. */
					ret=ftdi_write_data (&ftdic_a, &outbyte, 1);	/* Write length is 1 byte */
					if (ret<0){
						fprintf(stderr, "Error %d: ftdi_write_data to PortA failed: (%s)\n", ret, ftdi_get_error_string(&ftdic_a));
						exit (FTDI_ERROR_WRITEFAIL);
					}

					/* Sleep 0.125 sec */
					usleep (125000);

					/* All zeros */
					outbyte = 0x00;
					eprintf("    %02hhx\n", outbyte);

					/* Write the byte to PortA. */
					ret=ftdi_write_data (&ftdic_a, &outbyte, 1);	/* Write length is 1 byte */
					if (ret<0){
						fprintf(stderr, "Error %d: ftdi_write_data to PortA failed: (%s)\n", ret, ftdi_get_error_string(&ftdic_a));
						exit (FTDI_ERROR_WRITEFAIL);
					}

					/* Sleep 0.375 sec */
					usleep (375000);
				}

			}else{

				/* Begin test pattern */
				fprintf (stderr, "Starting to write test pattern to Port A\n");		/* Say this, even in non-debug mode */
				fprintf (stderr, "10Hz Running LEDs, from LSB -> MSB. PortA = 'Electrons'...\n");

				eprintf("    PortA\n    -----\n");
				for (i=0;i<128;i++){				/* Pattern lasts about 12 sec */
					outbyte=0x01 << (i%8);

					eprintf("    %02hhx\n", outbyte);

					/* Write the byte to PortA. */
					ret=ftdi_write_data (&ftdic_a, &outbyte, 1);	/* Write length is 1 byte */
					if (ret<0){
						fprintf(stderr, "Error %d: ftdi_write_data to PortA failed: (%s)\n", ret, ftdi_get_error_string(&ftdic_a));
						exit (FTDI_ERROR_WRITEFAIL);
					}

					/* Sleep 0.1 sec */
					usleep (100000);
				}

				outbyte=0x0a;  /* Identify the port as A */

				eprintf("\n    PortA\n    -----\n    %02hhx  \n\n", outbyte);

				/* Write the byte to PortA */
				ret=ftdi_write_data (&ftdic_a, &outbyte, 1);	/* Write length is 1 byte */
				if (ret<0){
					fprintf(stderr, "Error %d: ftdi_write_data to PortA failed: (%s)\n", ret, ftdi_get_error_string(&ftdic_a));
					exit (FTDI_ERROR_WRITEFAIL);
				}
			}

			/* Do NOT disable bitbang mode, or the last swritten bytes won't persist at the outputs. */
			eprintf("NOT disabling bitbang mode for PortA, so that the last bytes persist at the outputs.\n");

			eprintf("Closing PortA...\n");
			ret=ftdi_usb_close(&ftdic_a);	/* Only try to close if was open; else would segfault*/
			if (ret<0){
				fprintf(stderr, "Error %d: ftdi_usb_close failed for PortA: (%s)\n", ret, ftdi_get_error_string(&ftdic_a));
				exit (FTDI_ERROR_CLOSEFAIL);
			}
		}

		/* De-initialise ftdi contexts. (void) */
		ftdi_deinit(&ftdic_a);

		/* If we got here, it was successful */
		fprintf (stderr, "Finished test pattern.\n");   /* Say this, even in non-debug mode */
		exit (FTDI_EXIT_OK);
	}

	/* If we get here without having exited, something (such as a while(1) exiting) went wrong! */
	fprintf(stderr,"Error: reached end of program. How did we get here?\n");
	exit (FTDI_ERROR_GENERIC);
}
